#include "prjlibs-c/standards.h"
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/vt.h>

#include "skalibs/include/stddjb.h"
#include "skalibs/include/ttymodes.h"
#include "prjlibs-c/constants.h"
#include "fdtools.h"

char const* PROG="vc-lock";

static int fd;

static void release_vt(int sig) {
  (void)sig;
  ioctl(fd, VT_RELDISP, 0);
}

static void acquire_vt(int sig) {
  (void)sig;
  ioctl(fd, VT_RELDISP, VT_ACKACQ);
}

int main(int argc, char** argv) {
  struct vt_mode vtm, old_vtm;
  ttymodes modes, old_modes;
  sigset_t set, old_set;
  int scan;

  if (argc<3)
    strerr_die3x(100, usage, PROG, " {fd|/dev/ttyN} program [arg ...]");

  fd=scan_or_open(&scan, argv[1], O_RDWR|O_NOCTTY, 0);
  if (fd<0) {
    if (scan) strerr_die3x(100, PROG, ": descriptor malformed: ", argv[1]);
    else strerr_die5sys(errstat, PROG, err_unable, err_open, argv[1], ": ");
  }

  if (ioctl(fd, VT_GETMODE, &old_vtm)<0)
    strerr_die3sys(errstat, PROG, err_unable, "get VT mode: ");

  if (ttymodes_gett(fd, &old_modes)!=0)
    strerr_die4sys(100, PROG, err_unable, err_tcgetattr, ": ");

  sig_catch(SIGUSR1, release_vt);
  sig_catch(SIGUSR2, acquire_vt);
  sigemptyset(&set);
  sigaddset(&set, SIGUSR1);
  sigaddset(&set, SIGUSR2);
  if (sigprocmask(SIG_UNBLOCK, &set, &old_set)!=0)
    strerr_die4sys(100, PROG, err_unable, err_sigmask, ": ");

  modes=old_modes;
  modes.ti.c_iflag&=~BRKINT;
  modes.ti.c_iflag|= IGNBRK;
  modes.ti.c_lflag&=~ISIG;
  if (ttymodes_sett(fd, &modes)!=0)
    strerr_die4sys(100, PROG, err_unable, err_tcsetattr, ": ");

  vtm=old_vtm;
  vtm.mode=VT_PROCESS;
  vtm.relsig=SIGUSR1;
  vtm.acqsig=SIGUSR2;
  if (ioctl(fd, VT_SETMODE, &vtm)<0)
    strerr_die3sys(errstat, PROG, err_unable, "set VT mode: ");

  argv+=2;
  for (;;) {
    pid_t const pid=fork();
    if (pid<0) {
      strerr_warn4sys(PROG, err_unable, err_fork, ": ");
    } else if (pid==0) {
      sigprocmask(SIG_SETMASK, &old_set, null);
      pathexec((char const**)argv);
      strerr_die5sys(errstat, PROG, err_unable, err_exec, *argv, ": ");
    } else {
      int status;
      if (wait_pid(&status, pid)!=pid)
        strerr_warn4sys(PROG, err_unable, err_wait, ": ");
      else if (status==0) break;
    }
    deepsleep(1);
  }

  if (ioctl(fd, VT_SETMODE, &old_vtm)<0)
    strerr_die3sys(errstat, PROG, err_unable, "set VT mode");
  if (ttymodes_sett(fd, &old_modes)!=0)
    strerr_die4sys(100, PROG, err_unable, err_tcsetattr, ": ");

  _exit(0);
}
